/* mc.h -- minicache headers
  Copyright 1997-98 Michael Callahan
  Copyright 1999 Carnegie Mellon University
  This program is free software.  You may copy it according
  to the conditions of the GNU General Public License version 2.0
  or later; see the file COPYING in the source directory.
*/


#define WANTVXDWRAPS
#include <basedef.h>
#include <vmm.h>
#include <ios.h>
#include <ifs.h>
#include <vwin32.h>
#include <vxdwraps.h>

#include <winerror.h>

#include "ifsmgr.h"
#include "iosdcls.h"

#include "ipc.h"

#include "cnode.h"

#include "errno.h"

#pragma warning(disable: 4102)

/* init.c */

extern int inUseCount;
BOOL OnDeviceInit();

/* devio.c */

DWORD OnW32DeviceIoControl(HVM, PDIOCPARAMETERS);

extern int calls[10];

/* vol.c */

extern struct cnode *root;
int ConnectNetResource (pioreq);
int DisconnectResource (pioreq);

int ErrorStub (pioreq);

int FileAttributes (pioreq);
int DeleteFile (pioreq);
int FlushVolume (pioreq);
int Dir (pioreq);
int GetDiskInfo (pioreq);
int QueryResourceInfo (pioreq);
int RenameFile (pioreq);
int UNCNamedPipeRequest (pioreq);

extern UINT uiNetId;
extern UINT uiProviderId;

/* misc.c */

int errno2res (int errno);

void logucode (PWORD);
void logmsg (char *);
void printf (char *, ...);

struct request *NewRequest(int);
void FreeRequest (struct request *);

struct cnode *NewCnode(void);
void FreeCnode (struct cnode *);

void fmt_dword (char *, DWORD);
void print_pir (pioreq);
void print_ppath (ParsedPath *);

PUSHORT to_ucode(PCHAR name);
PCHAR to_BCS(PUSHORT name, int length);

/* file.c */

int OpenFile (pioreq);
int ReadFile (pioreq);
int WriteFile (pioreq);
int FileSeek (pioreq);
int CommitFile (pioreq);
int CloseFile (pioreq);
int LockFile (pioreq);
int FileDateTime (pioreq);
int NamedPipeRequest (pioreq);
int NamedPipeHandleInfo (pioreq);
int EnumerateHandle (pioreq);

/* find.c */

int SearchFile (pioreq);
int FindFirstFile (pioreq);
int FindNextFile (pioreq);
int FindFileClose (pioreq);
int EnumerateFind (pioreq);

/* ipc.c */

int initialize_ipc();
int process_rpc (struct request *);
int insert_upcall (struct request *req);
struct request *get_pending();
struct request *find_request(long);
void signal_request(struct request *);
int set_R3_event(DWORD);
void clear_resource_requests(DWORD rh);
void heartbeat();
DWORD do_poll(PVOID, PDWORD);
void do_close(PVOID);

/* readdir.c */

#define READDIR_BUFSIZE 2048

typedef struct READDIR {
	ViceFid		VFid;
	DWORD		hdl;
	DWORD		filesize;
	DWORD		pos;
	DWORD		bufstart;
	DWORD		bufend;
	char		buf[READDIR_BUFSIZE];
} READDIR, *PREADDIR;

int opendir (ViceFid VFid, PREADDIR *p);
int readdir (PREADDIR p, PCHAR q, PDWORD plen);
void closedir(PREADDIR p);

/* cfs.c */

int openVFid (ViceFid VFid, int flags, WORD access, WORD attr, BYTE action, 
			  struct request *inreq, PDWORD phdl, PDWORD ptakenaction,
			  PWORD pattr);
int pathelement_lookup (PathElement *pe, struct cnode *parent, struct cnode **dir);
int open_namei (ParsedPath *, struct cnode **);
int namei (ParsedPath *, USHORT *, struct cnode **);
int ascii_namei (PCHAR, struct cnode **);
struct cnode * iget (ViceFid fid);
void iput (struct cnode *);
int lookup (ViceFid dirfid, PCHAR name, ViceFid *fid, DWORD *type, int flags);
int lookup_ucode (ViceFid dirfid, USHORT *name, ViceFid *fid, DWORD *type);
int getattr (ViceFid fid, struct coda_vattr *vattr);
int coda_getattr (struct cnode *);
int coda_setattr (struct cnode *);

/* upcall.c */

int venus_remove(struct cnode *, pioreq );
int venus_mkdir(struct cnode *, pioreq );
int venus_rmdir(struct cnode *, pioreq );
int venus_root(ViceFid **);
int venus_close(ViceFid, DWORD );
int venus_create(struct cnode *, pioreq , ViceFid *);
int venus_pioctl(ViceFid, struct PioctlData *, PCHAR *, DWORD, PDWORD);
int venus_lookup(ViceFid, string_t, ViceFid *, int *, int, int, int );
int venus_open_by_path(ViceFid, char*, struct request *, int);
int venus_getattr(ViceFid fid, struct coda_vattr *vattr);
int venus_setattr(ViceFid fid, struct coda_vattr vattr);
int venus_rename(struct cnode *cn1, struct cnode *cn2, pioreq pir);

/* convert.c */

typedef struct CTENTRY {
	CHAR			shortname[13];
	CHAR			longname[MAX_PATH+1];
	struct CTENTRY	*next;
} CTENTRY, *PCTENTRY;

void convert_table_add(PCTENTRY head, PCHAR *s, PCHAR l);
void convert_table_init(PCTENTRY head);
void convert_table_flush(PCTENTRY head);
