/* upcalls.c
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/
#include "mc.h"

#include "string.h"

#pragma intrinsic(strlen, strcpy, memset, memcpy)

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

#define INSIZE(tag) sizeof(struct coda_ ## tag ## _in)
#define OUTSIZE(tag) sizeof(struct coda_ ## tag ## _out)
#define SIZE(tag)  max(INSIZE(tag), OUTSIZE(tag))

/* put the current process credentials in the cred */
void coda_load_creds(struct coda_cred *cred)
{
        cred->cr_uid = 500;
        cred->cr_euid = 500;
        cred->cr_suid = 500;
        cred->cr_fsuid = 500;

        cred->cr_groupid = 500;
        cred->cr_egid = 500;
        cred->cr_sgid = 500;
        cred->cr_fsgid = 500;
}

#define UPARG(op, length)\
do {\
	req = NewRequest(length);\
	if (!req) {\
		res = ERROR_NOT_ENOUGH_MEMORY;\
		return res;\
	}\
    req->r.in.ih.opcode = (op);\
	coda_load_creds(&(req->r.in.ih.cred));\
} while (0)

int venus_remove(struct cnode *cn, pioreq pir){
	struct request *req = NULL;
	PCHAR p;
	int res = 0, len, i;

	for (len = 0; pir->ir_uFName[len]; len++)
		;

/*	req = NewRequest(len + 1);
	if (!req) {
		res = ERROR_NOT_ENOUGH_MEMORY;
		return res;
	}

	req->r.in.ih.opcode = CODA_REMOVE;*/

	UPARG(CODA_REMOVE, len+1);

	req->r.in.coda_remove.VFid = cn->c_fid;
	p = (char *) (&req->r.in.coda_remove.name + 1);
	req->r.in.coda_remove.name = (p - (char *) &req->r.in);
	for (i = 0; pir->ir_uFName[i]; i++)
		p[i] = (CHAR) pir->ir_uFName[i];
	p[i] = '\0';

	res = process_rpc(req);
	if (res) {
		printf ("remove process_rpc failed %x", res);
		if (req) FreeRequest (req);
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		printf ("remove rpc failed %x", res);
		res = errno2res(res);
		if (req) FreeRequest (req);		
		return res;
	}
	return res;
}

int venus_mkdir(struct cnode *cn, pioreq pir){
	struct request *req = NULL;
	int len, i;
	int res = ERROR_INVALID_FUNCTION;
	PCHAR p;

	for (len = 0; pir->ir_uFName[len]; len++)
			;
/*	req = NewRequest(len + 1);
	if (!req) {
		res = ERROR_NOT_ENOUGH_MEMORY;
		return res;
	}

	req->r.in.ih.opcode = CODA_MKDIR;*/
	UPARG(CODA_MKDIR, len+1);

	req->r.in.coda_mkdir.VFid = cn->c_fid;
	req->r.in.coda_mkdir.attr.va_mode = 0666;
	p = (char *) (&req->r.in.coda_mkdir.name + 1);
	req->r.in.coda_mkdir.name = (p - (char *) &req->r.in);
	for (i = 0; pir->ir_uFName[i]; i++)
		p[i] = (CHAR) pir->ir_uFName[i];
	p[i] = '\0';

	res = process_rpc(req);
	if (res) {
		printf ("mkdir process_rpc failed %x", res);
		if (req) FreeRequest (req);		
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		printf ("mkdir rpc failed %x", res);
		res = errno2res(res);
		if (req) FreeRequest (req);		
		return res;
	}
	return res;
}

int venus_rmdir(struct cnode *cn, pioreq pir){
	struct request *req = NULL;
	int len, i;
	int res = ERROR_INVALID_FUNCTION;
	PCHAR p;

	for (len = 0; pir->ir_uFName[len]; len++)
		;
/*	req = NewRequest(len + 1);
	if (!req) {
		res = ERROR_NOT_ENOUGH_MEMORY;
		return res;
	}

	req->r.in.ih.opcode = CODA_RMDIR;*/
	UPARG(CODA_RMDIR, len+1);

	req->r.in.coda_rmdir.VFid = cn->c_fid;
	p = (char *) (&req->r.in.coda_rmdir.name + 1);
	req->r.in.coda_rmdir.name = (p - (char *) &req->r.in);
	for (i = 0; pir->ir_uFName[i]; i++)
		p[i] = (CHAR) pir->ir_uFName[i];
	p[i] = '\0';

	res = process_rpc(req);
	if (res) {
		printf ("rmdir process_rpc failed %x", res);
		if (req) FreeRequest (req);		
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		printf ("rmdir rpc failed %x", res);
		res = errno2res(res);
		if (req) FreeRequest (req);		
		return res;
	}
	return res;
}

int venus_root(ViceFid **fid){
	struct request *req = NULL;
	int res = 0;

/*	req = NewRequest(0);
	if (!req) {
		res = ERROR_NOT_ENOUGH_MEMORY;
		return res;
	}

	req->r.in.ih.opcode = CODA_ROOT;*/
	UPARG(CODA_ROOT, 0);

	res = process_rpc(req);
	if (res) {
		logmsg ("CODA_ROOT rpc failed");
		if (req) FreeRequest (req);		
		return res;
	}
	if (req->r.out.oh.result) {
		logmsg ("CODA_ROOT nonzero venus result");
		res = errno2res (req->r.out.oh.result);
		if (req) FreeRequest (req);		
		return res;
	}

	printf ("ROOT VFID: %x %x %x", req->r.out.coda_root.VFid.Volume,
		req->r.out.coda_root.VFid.Vnode, req->r.out.coda_root.VFid.Unique);

	*fid = &req->r.out.coda_root.VFid;
	return res;
}

int venus_close(ViceFid VFid, DWORD flags){
	struct request *req;
	int res = ERROR_ACCESS_DENIED;
	
	UPARG(CODA_CLOSE, 0);

//	req = NewRequest(0);

//	if (req){
//		req->r.in.ih.opcode = CODA_CLOSE;
		req->r.in.coda_close.VFid = VFid;
		req->r.in.coda_close.flags = flags;
		res = process_rpc(req);
		if (res) 
			printf ("CODA_CLOSE process_rpc failed: %x", res);

		res = req->r.out.oh.result;
		if (res) 
			printf ("CODA_CLOSE rpc failed: %x", res);

		if (req) 
			FreeRequest(req);
//	} else 
//		logmsg ("venus_close: out of memory");
	return res;
}

int venus_create(struct cnode *cn, pioreq pir, ViceFid *fid){
	struct request *req = NULL;
	PCHAR p;
	int i, res;

/*	req = NewRequest(MAX_PATH + 1);
	if (!req) {
		res = ERROR_NOT_ENOUGH_MEMORY;
		return res;
	}*/
	UPARG(CODA_CREATE, MAX_PATH+1);

	if (pir->ir_options & ACTION_TRUNCATE)
		req->r.in.coda_create.attr.va_size = 0;

//	req->r.in.ih.opcode = CODA_CREATE;
	req->r.in.coda_create.VFid = cn->c_fid;
	req->r.in.coda_create.attr.va_mode = 0666;
	req->r.in.coda_create.excl = (pir->ir_options & ACTION_TRUNCATE) ? 0 : 1;
	req->r.in.coda_create.mode = 0666;
	p = (char *) (&req->r.in.coda_create.name + 1);
	req->r.in.coda_create.name = (p - (char *) &req->r.in);
	for (i = 0; pir->ir_uFName[i]; i++)
		p[i] = (CHAR) pir->ir_uFName[i];
	p[i] = '\0';

	logmsg ((char *) (&req->r.in.coda_create.name + 1));
	
	res = process_rpc(req);
	if (res) {
		logmsg ("create process_rpc failed");
		if (req) FreeRequest(req);
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		logmsg ("create failed");
		res = errno2res(res);
		if (req) FreeRequest(req);
		return res;
	}

	fid->Volume = req->r.out.coda_lookup.VFid.Volume;
	fid->Vnode  = req->r.out.coda_lookup.VFid.Vnode;
	fid->Unique = req->r.out.coda_lookup.VFid.Unique;

	return res;
}

typedef unsigned char RPC2_EncryptionKey[8];
typedef unsigned char EncryptedSecretToken[56];
typedef struct{
	long AuthHandle;
	RPC2_EncryptionKey HandShakeKey;
	long ViceId;
	long BeginTimestamp;
	long EndTimestamp;
} ClearToken;
typedef struct {
    int			    sTokenSize;
    EncryptedSecretToken    stoken;
    int			    cTokenSize;
    ClearToken		    ctoken;
} venusbuff;

int venus_pioctl(ViceFid dirfid, struct PioctlData *pPID, PCHAR *out, DWORD out_size, PDWORD ret_count){
	
	struct request *req = NULL;
	DWORD offset, size, res = 0, iocsize;
	venusbuff *vb;

	size = pPID->vi.in_size;
	if (out_size > size)
		size = out_size;

	UPARG(CODA_IOCTL, size);
	offset = INSIZE(ioctl);

    /* the cmd field was mutated by increasing its size field to
     * reflect the path and follow args. We need to subtract that
     * out before sending the command to Venus.  */
    req->r.in.coda_ioctl.cmd = (pPID->cmd & ~(PIOCPARM_MASK << 16));	
    iocsize = ((pPID->cmd >> 16) & PIOCPARM_MASK) - sizeof(char *) - sizeof(int);
    req->r.in.coda_ioctl.cmd |= (iocsize & PIOCPARM_MASK) <<	16;	

	req->r.in.coda_ioctl.VFid = dirfid;
	req->r.in.coda_ioctl.len = pPID->vi.in_size;
	req->r.in.coda_ioctl.data = (PCHAR)offset;

	vb = (venusbuff *)pPID->vi.in;
	printf ("venusbuff: ctoken.EndTimestamp: %x, size: %x, offset: %x", vb->ctoken.EndTimestamp, size, offset);
	memcpy ((PCHAR)&(req->r.in) + offset, pPID->vi.in, pPID->vi.in_size);

	logmsg("--Pioctl: before process_rpc--");
	res = process_rpc(req);
	if (res) {
		printf ("CODA_IOCTL rpc failed %x", res);
		if (req) FreeRequest(req);
		return res;
	}
	if (req->r.out.oh.result) {
		printf ("CODA_IOCTL nonzero venus result %x", req->r.out.oh.result);
		res = req->r.out.oh.result;
//		if (req) FreeRequest(req);
//		return (errno2res (req->r.out.oh.result));
	}

	if (out_size != 0){
		if ((unsigned)req->r.out.coda_ioctl.len > out_size) {
			logmsg("--Pioctl: output bigger than buffer--");
			if (req) FreeRequest(req);
			return ERROR_INVALID_PARAMETER;

		}
		memcpy(*out, ((PCHAR) &req->r.out) + (DWORD) req->r.out.coda_ioctl.data,
			req->r.out.coda_ioctl.len);
		*ret_count = req->r.out.coda_ioctl.len;
	}

	if (req) FreeRequest(req);
	return res;
}

int venus_lookup(ViceFid dirfid, string_t uFName, ViceFid *fid,
				 int *vtype, int flags, int extrasize, int file){
	struct request *req = NULL;
	int offset, res = 0;


	UPARG(CODA_LOOKUP, extrasize);

	offset = INSIZE(lookup);
	req->r.in.coda_lookup.VFid = dirfid;
	req->r.in.coda_lookup.name = offset;
	req->r.in.coda_lookup.flags = flags;
	/* send Venus a null terminated string */
	UniToBCS((char *)(&(req->r.in)) + offset, uFName, extrasize*2-2, extrasize-1, BCS_OEM);
    *((char *)&(req->r.in) + offset + extrasize) = '\0';

	/* first, if the request flags allow it, we try to open existing file */
	logmsg ((char *) (&(req->r.in)) + offset);

	res = process_rpc(req);
	if (res) {
		logmsg ("lookup process_rpc failed");
		if (req) FreeRequest(req);
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		res = errno2res (res);
		printf ("lookup error %x", res);
		if (req) FreeRequest(req);
		return res;
	}
	
	if (file && req->r.out.coda_lookup.vtype != C_VREG) {
		/* oops, found something, but it's not a file! */
		res = ERROR_ACCESS_DENIED;
		if (req) FreeRequest(req);
		return res;
	}

	fid->Volume = req->r.out.coda_lookup.VFid.Volume;
	fid->Vnode  = req->r.out.coda_lookup.VFid.Vnode;
	fid->Unique = req->r.out.coda_lookup.VFid.Unique;
	*vtype = req->r.out.coda_lookup.vtype;

	if (req) FreeRequest(req);
	return res;
	
}

int venus_open_by_path(ViceFid VFid, char * path, struct request *inreq, int flags){
	struct request *req = inreq;
	int res;

/*	if (!req && !(req = NewRequest(MAX_PATH + 1))) {
		logmsg ("venus_open_by_path: out of memory\n");
		return ERROR_NOT_ENOUGH_MEMORY;
	}*/

	UPARG(CODA_OPEN_BY_PATH, MAX_PATH+1);

	printf ("VFID: %x %x %x", VFid.Volume,VFid.Vnode, VFid.Unique);
//	req->r.in.ih.opcode = CODA_OPEN_BY_PATH;
	req->r.in.coda_open_by_path.VFid = VFid;
	req->r.in.coda_open_by_path.flags = flags;

	res = process_rpc(req);
	if (res) {
		printf ("CODA_OPEN_BY_PATH process_rpc failed");
		if (!inreq)	FreeRequest(req);
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		printf ("CODA_OPEN_BY_PATH rpc failed");
		res = errno2res (res);
		if (!inreq)	FreeRequest(req);
		return res;
	}

	/* get path */
	strcpy (path, ((char *) &req->r.out) + req->r.out.coda_open_by_path.path);
	logmsg (path);

	if (!inreq) FreeRequest(req);
	return res;
}

int venus_getattr(ViceFid fid, struct coda_vattr *vattr){
	struct request *req = NewRequest(0);
	int res;

/*	if (!req)
		return ERROR_NOT_ENOUGH_MEMORY;

	req->r.in.ih.opcode = CODA_GETATTR;*/
	UPARG(CODA_GETATTR, 0);

	req->r.in.coda_getattr.VFid = fid;

	res = process_rpc(req);
	if (res) {
		printf ("getattr process_rpc failed %x", res);
		if (req) FreeRequest(req);
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		printf ("getattr rpc failed %x", res);
		res = errno2res(res);
		if (req) FreeRequest(req);
		return res;
	}

	*vattr = req->r.out.coda_getattr.attr;

error:
	if (req) FreeRequest(req);
	return res;
}

int venus_setattr(ViceFid fid, struct coda_vattr vattr){
	struct request *req = NULL;
	int res = 0;

/*	req = NewRequest(0);
	if (!req) {
		res = ERROR_NOT_ENOUGH_MEMORY;
		return res;
	}*/

	vattr.va_nlink     = -1;
	vattr.va_blocksize = -1;
	vattr.va_rdev      = -1;
	vattr.va_flags     = -1;
	vattr.va_fileid    = -1;
	vattr.va_bytes     = -1;
	vattr.va_size      = -1;

	UPARG(CODA_SETATTR, 0);

//	req->r.in.ih.opcode = CODA_SETATTR;
	req->r.in.coda_setattr.VFid = fid;
	req->r.in.coda_setattr.attr = vattr;

	res = process_rpc(req);
	if (res) {
		printf ("setattr process_rpc failed %x", res);
		if (req) FreeRequest(req);
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		printf ("setattr rpc failed %x", res);
		res = errno2res(res);
		if (req) FreeRequest(req);
		return res;
	}
	if (req) FreeRequest(req);
	return res;
}

int venus_rename(struct cnode *cn1, struct cnode *cn2, pioreq pir){

	struct request *req = NULL;
	int res = 0, i, len;
	PathElement* old_name = NULL;
	PCHAR p;

	old_name = IFSLastElement( pir->ir_ppath );

	for (len = 0; pir->ir_uFName[len]; len++)
		;

/*	req = NewRequest(len + 1);
	if (!req) {
		res = ERROR_NOT_ENOUGH_MEMORY;
		return res;
	}

	req->r.in.ih.opcode = CODA_RENAME;*/
	UPARG(CODA_RENAME, len+1);

	req->r.in.coda_rename.sourceFid = cn1->c_fid;
	req->r.in.coda_rename.destFid = cn2->c_fid;

	p = (char *) (&req->r.in.coda_rename.destname + 1);
	req->r.in.coda_rename.destname = (p - (char *) &req->r.in);
	for (i = 0; pir->ir_uFName[i]; i++)
		p[i] = (CHAR) pir->ir_uFName[i];
	p[i] = '\0';

	p += i+1;
	req->r.in.coda_rename.srcname = (p - (char *) &req->r.in);
	for (i = 0; old_name->pe_unichars[i]; i++) {
		p[i] = (CHAR) old_name->pe_unichars[i];
		if (p[i] >= 'A' && p[i] <= 'Z')
			p[i] += 'a' - 'A';
	}
	p[i] = '\0';

	res = process_rpc(req);
	if (res) {
		printf ("rename process_rpc failed %x", res);
		if (req) FreeRequest (req);
		return res;
	}
	res = req->r.out.oh.result;
	if (res) {
		printf ("rename rpc failed %x", res);
		res = errno2res(res);
		if (req) FreeRequest (req);
		return res;
	}

	if (req) FreeRequest (req);
	return res;
}
