/* vol.c -- Windows95 IFS routines for volume manipulation
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/


#include "mc.h"

#include "string.h"
#pragma intrinsic(strlen, strcpy, memset, memcpy)

#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

static struct volfunc vf = {
	IFS_VERSION,
	IFS_REVISION,
	NUM_VOLFUNC,

	DeleteFile,				/* DeleteFile */
	Dir,					/* Dir */
	FileAttributes,			/* FileAttributes */
	FlushVolume,			/* FlushVolume */
	GetDiskInfo,			/* GetDiskInfo */

	OpenFile,				/* OpenFile */
	RenameFile,				/* RenameFile */
	SearchFile,				/* SearchFile */
	QueryResourceInfo,		/* QueryResourceInfo */
	DisconnectResource,		/* DisconnectResource */

	UNCNamedPipeRequest,	/* UNCNamedPipeReq */
	ErrorStub,				/* Ioctl16Drive */
	ErrorStub,				/* GetDiskParms */
	FindFirstFile,			/* FindFirstFile */
	ErrorStub				/* DirectDiskIO */
};

UINT uiNetId = 0x001b0000;
UINT uiProviderId = 0;

struct cnode *root = NULL;

int ConnectNetResource (pioreq pir)
{
	ViceFid *VFid;
	struct cnode *cn = NULL;
	int res = 0; 

	logmsg ("--ConnectNetResource--");

	if ((pir->ir_flags != RESTYPE_WILD) &&
		(pir->ir_flags != RESTYPE_DISK))
		return pir->ir_error = ERROR_NOT_SUPPORTED;

	/* get root VFid */

	res = venus_root(&VFid);

	pir->ir_flags = RESTYPE_DISK;
	pir->ir_pathSkip = 2;
	pir->ir_vfunc = &vf;

	cn = iget(*VFid);
	if (!cn) {
		res = ERROR_NOT_ENOUGH_MEMORY;
		return pir->ir_error = res;
	}
	root = cn;

	inUseCount++;

	if (res != 0 && cn) FreeCnode(cn);

	logmsg ("--end of ConnectNetResource--");
	return pir->ir_error = res;
}


int DisconnectResource (pioreq pir)
{
	logmsg ("--disconnect--");
	clear_resource_requests ((DWORD) pir->ir_rh);

	inUseCount--;
	iput(root);
	root = NULL;
	logmsg ("--end of disconnect--");
	return pir->ir_error = 0;
}


int ErrorStub (pioreq pir)
{
	logmsg ("--errorstub--");
	logmsg ("--end of errorstub--");
	return pir->ir_error = ERROR_INVALID_FUNCTION;
}


int FileAttributes (pioreq pir)
{
	struct cnode *cn = NULL;
	int res = 0;

	printf ("--fileattrib-- (subfunc %x)", pir->ir_flags);
	logucode (pir->ir_upath);

	res = namei(pir->ir_ppath, pir->ir_uFName, &cn);
	if (res)
		goto error;

	if (pir->ir_flags == GET_ATTRIBUTES) {
		pir->ir_attr = (cn->c_vattr.va_type == C_VDIR) ?
					FILE_ATTRIBUTE_DIRECTORY : 0;
		res = 0;
		goto error;
	}

	if (pir->ir_flags == GET_ATTRIB_FIRST_CLUST) {
		res = ERROR_INVALID_FUNCTION;
		goto error;
	}

	/* otherwise, it is one of the date/time functions:
	   get current attributes */
	res = coda_getattr(cn);
	if (res)
		goto error;

	switch (pir->ir_flags) {
	case GET_ATTRIB_COMP_FILESIZE:
		logmsg("--fileattrib: GET_ATTRIB_COMP_FILESIZE--");
		pir->ir_size = (unsigned long)cn->c_vattr.va_size;
		break;

	case SET_ATTRIB_MODIFY_DATETIME:
		logmsg("--fileattrib: SET_ATTRIB_MODIFY_DATETIME--");
		cn->c_vattr.va_mtime.ts_sec = IFSMgr_DosToNetTime(pir->ir_dostime);
		cn->c_vattr.va_mtime.ts_nsec = 0;
		break;

	case GET_ATTRIB_MODIFY_DATETIME:
		logmsg("--fileattrib: GET_ATTRIB_MODIFY_DATETIME--");
		pir->ir_dostime = IFSMgr_NetToDosTime(cn->c_vattr.va_mtime.ts_sec);
		break;

	case SET_ATTRIB_LAST_ACCESS_DATETIME:
		logmsg("--fileattrib: SET_ATTRIB_LAST_ACCESS_DATETIME--");
		cn->c_vattr.va_atime.ts_sec = IFSMgr_DosToNetTime(pir->ir_dostime);
		cn->c_vattr.va_atime.ts_nsec = 0;
		break;

	case GET_ATTRIB_LAST_ACCESS_DATETIME:
		logmsg("--fileattrib: GET_ATTRIB_LAST_ACCESS_DATETIME--");
		pir->ir_dostime = IFSMgr_NetToDosTime(cn->c_vattr.va_atime.ts_sec);
		break;

	case SET_ATTRIB_CREATION_DATETIME:
		logmsg("--fileattrib: SET_ATTRIB_CREATION_DATETIME--");
		cn->c_vattr.va_ctime.ts_sec = IFSMgr_DosToNetTime(pir->ir_dostime);
		cn->c_vattr.va_ctime.ts_nsec = pir->ir_options * 10000000;
		break;

	case GET_ATTRIB_CREATION_DATETIME:
		logmsg("--fileattrib: GET_ATTRIB_CREATION_DATETIME--");
		pir->ir_dostime = IFSMgr_NetToDosTime(cn->c_vattr.va_ctime.ts_sec);
		pir->ir_options = cn->c_vattr.va_ctime.ts_nsec / 10000000;
		break;
	}

	if (pir->ir_flags == SET_ATTRIB_MODIFY_DATETIME ||
		pir->ir_flags == SET_ATTRIB_LAST_ACCESS_DATETIME ||
		pir->ir_flags == SET_ATTRIB_CREATION_DATETIME)
		res = coda_setattr(cn);

error:
	printf ("--end of fileattrib (%x)--", res);
	if (cn) iput(cn);
	return pir->ir_error = res;
}


int DeleteFile (pioreq pir)
{
	struct cnode *cn = NULL;
	int res;

	logmsg ("--deletefile--");
	/* according to docs, returning INVALID_FUNCTION on a wildcard
		request will cause the IFSMgr to retry with a sequence of
		non-wildcard requests */
	if (pir->ir_attr & FILE_FLAG_WILDCARDS)
		return ERROR_INVALID_FUNCTION;

	res = open_namei(pir->ir_ppath, &cn);

	if (!res)
		res = venus_remove(cn, pir);

	if (cn) iput (cn);

	printf ("delete returning %x", res);
	return pir->ir_error = res;

}


int FlushVolume (pioreq pir)
{
	logmsg ("--flushvolume--");
	logmsg ("--end of flushvolume--");
	return pir->ir_error = 0;
}


int Dir (pioreq pir)
{
	int res = ERROR_INVALID_FUNCTION;
	struct cnode *cn = NULL;
	int i = 1;

	printf ("--dir-- (subfunc %x)", pir->ir_flags);

	if (pir->ir_flags == QUERYLONG_DIR ) {
		logmsg ("--querylong_dir--");
		logucode((PWORD)pir->ir_ppath2);
		goto error;
	}

	if (pir->ir_flags == QUERY83_DIR){
		logmsg ("--query83_dir--");
		logucode((PWORD)pir->ir_ppath2);
		goto error;
	}

	if (pir->ir_flags == CHECK_DIR) {
		logmsg ("--check dir--");

		if (pir->ir_uFName[0] == 0){
			res = 0;
			goto error;
		}

        if (IFSIsRoot(pir->ir_ppath)) {
            res = 0;
			goto error;
		}
		
		logucode((PWORD)pir->ir_ppath);
		logucode((PWORD)pir->ir_upath);
		res = namei(pir->ir_ppath, pir->ir_uFName, &cn);
		if (res==ERROR_FILE_NOT_FOUND){
			res = ERROR_PATH_NOT_FOUND;
			goto error;
		}
		if (cn->c_vattr.va_type != C_VDIR)
			res = ERROR_PATH_NOT_FOUND; 
		goto error;
	}

	if (pir->ir_flags == CREATE_DIR) {
		logmsg ("--create dir--");

		res = open_namei(pir->ir_ppath, &cn);

		if (!res)
			res = venus_mkdir(cn, pir);
	}

	if (pir->ir_flags == DELETE_DIR) {
		logmsg ("--delete dir--");

		if (pir->ir_attr & FILE_FLAG_WILDCARDS)
			return ERROR_INVALID_FUNCTION;
		
		res = open_namei(pir->ir_ppath, &cn);

		if (!res)
			res = venus_rmdir(cn, pir);
	}

error:
	if (cn) iput(cn);

	printf ("--end of dir: %x--", res);
	return pir->ir_error = res;
}



int GetDiskInfo (pioreq pir)
{
	logmsg ("--getdiskinfo--");

	pir->ir_length = 512;			/* bytespersector */
	pir->ir_size = 10000;			/* totalallocunits */
	pir->ir_sectors = 2;			/* sectorsperallocunit */
	pir->ir_numfree = 5000;			/* numfreeallocunits */

	logmsg ("--end of getdiskinfo--");
	return pir->ir_error = 0;
}

static USHORT resourcename[] = { 32, 16,
   12,'\\', 'C', 'O', 'D', 'A',
   16, 'C', 'L', 'U', 'S', 'T', 'E', 'R',
	0 };

int QueryResourceInfo (pioreq pir)
{
	DWORD len;
	logmsg ("--queryresourceinfo--");

	switch (pir->ir_options) {
	case 1:
		pir->ir_options = RESSTAT_OK;
		pir->ir_flags = RESTYPE_DISK;
		/* fallthrough */
	case 0:
		/* URGH! hooking Microsoft's FSDs reveals what we're supposed to do
		   here -- pity the documentation didn't */
		memcpy ((PCHAR) pir->ir_ppath, (PCHAR) resourcename, sizeof(resourcename));
		break;

	case 2:
	/* note: be careful to use pir->ir_length before overwriting it */
		printf ("level 2: data %x len %x", pir->ir_data, pir->ir_length);
		if (pir->ir_length) {
			len = 4;
			if (len > pir->ir_length - 1)
				len = pir->ir_length - 1;
			memcpy((unsigned char *) pir->ir_data, "CODA", 
				len);
			((PCHAR) pir->ir_data) [len] = '\0';
		}
		pir->ir_length = 0x010400ff;
		pir->ir_options = FS_CASE_IS_PRESERVED | FS_VOL_SUPPORTS_LONG_NAMES;
		pir->ir_pos = 512;
		break;

	default:
		logmsg ("QUERYRESOURCEINFO: unknown level");
	}

	logmsg ("--end of queryresourceinfo--");
	return pir->ir_error = 0;
}

int RenameFile (pioreq pir)
{
	struct cnode *cn1  = NULL;
	struct cnode *cn2 = NULL;
	int res = ERROR_INVALID_FUNCTION;
	int len, dummy2;
	ViceFid dummy1;


	logmsg ("--renamefile--");

	if (pir->ir_attr & FILE_FLAG_WILDCARDS || pir->ir_attr2 & FILE_FLAG_WILDCARDS)
		return ERROR_INVALID_FUNCTION;

	res = open_namei(pir->ir_ppath, &cn1);
	if (res)
		goto error;

	res = open_namei(pir->ir_ppath2, &cn2);
	if (res)
		goto error;
	
	for (len = 0; pir->ir_uFName[len]; len++)
		;
	len++;
	res = venus_lookup(cn2->c_fid, pir->ir_uFName, &dummy1, &dummy2, CLU_CASE_SENSITIVE, len, 0);  
	if (res == ERROR_FILE_NOT_FOUND) {
		res = venus_rename(cn1, cn2, pir);
	}
	else if (!res) res = ERROR_FILE_EXISTS;

error:
	if (cn1) iput (cn1);
	if (cn2) iput (cn2);

	printf ("--end of rename returning %x--", res);
	return pir->ir_error = res;
}

int UNCNamedPipeRequest (pioreq pir)
{
	printf ("--- failing UNCNamedPipeRequest %x", pir->ir_flags);
	return pir->ir_error = ERROR_INVALID_FUNCTION;
}