/*  
  Copyright 1997-98 Michael Callahan
  Copyright 1999 Carnegie Mellon University
  This program is free software.  You may copy it according
  to the conditions of the GNU General Public License version 2.0
  or later; see the file COPYING in the source directory.
*/

#include "sock.h"


int
sys_connect (HVM hVM, PCRS pRegs, PSOCKET pSock) 
{
	TDI_REQUEST r;
	PTRANSPORT_ADDRESS pTA = MakeTransportAddress(pRegs->Client_EBX, (WORD) pRegs->Client_EDX);
	PCALLBACKMEMO pCBM = NewCallback();
	PTDI_CONNECTION_INFORMATION pTCIRequest = _HeapAllocate(
				sizeof(TDI_CONNECTION_INFORMATION), HEAPZEROINIT);
	PTDI_CONNECTION_INFORMATION pTCIReturn = _HeapAllocate(
				sizeof(TDI_CONNECTION_INFORMATION), HEAPZEROINIT);
	int res = 0;
	TDI_STATUS err;

	if (!pTA || !pCBM || !pTCIRequest || !pTCIReturn) {
		res = -SOCKERR_NO_MEMORY;
		goto error;
	}

	if (! (pSock->s_flags & SOCKFL_BOUND)) {
		PTRANSPORT_ADDRESS pTA2 = MakeTransportAddress(0, 0);
		if (!pTA2) {
			res = -SOCKERR_NO_MEMORY;
			goto error;
		}

		res = do_bind(hVM, pRegs, pSock, pTA2);
		FreeTransportAddress(pTA2);
		if (res)
			goto error;
	}

	pSock->s_conhdl = makeconnection(pSock->s_addrobj->addr_hdl, pSock);
	if (!pSock->s_conhdl) {
		res = -SOCKERR_INTERNAL;
		goto error;
	}

	pTCIRequest->RemoteAddressLength = IP_TRANSPORT_ADDRESS_LENGTH;
	pTCIRequest->RemoteAddress = (PVOID) pTA;

	r.Handle.ConnectionContext = pSock->s_conhdl;
	r.RequestNotifyObject = &SignalCallback;
	r.RequestContext = (PVOID) pCBM;
	err = pDT->TdiConnectEntry (&r, NULL, pTCIRequest, pTCIReturn);
	if (err == TDI_PENDING)
		err = WaitCallback(pCBM);
	if (err == TDI_SUCCESS) {
		pSock->s_flags |= SOCKFL_CONNECTED;
		pSock->s_remoteport = (WORD) pRegs->Client_EDX;
		pSock->s_remoteip = pRegs->Client_EBX;
	} else {
		r.Handle.ConnectionContext = pSock->s_conhdl;
		r.RequestNotifyObject = &SignalCallback;
		r.RequestContext = (PVOID) pCBM;
		pDT->TdiCloseConnectionEntry (&r);
		pSock->s_conhdl = NULL;
	}
	_HeapFree(pTCIRequest, 0);
	_HeapFree(pTCIReturn, 0);
	FreeCallback(pCBM);
	FreeTransportAddress(pTA);
	return err;

error:
	if (pTCIRequest) _HeapFree(pTCIRequest, 0);
	if (pTCIReturn) _HeapFree(pTCIReturn, 0);
	if (pCBM) FreeCallback(pCBM);
	if (pTA) FreeTransportAddress(pTA);
	return res;
}

HANDLE
makeconnection (HANDLE addr, PSOCKET pSock)
{
	PCALLBACKMEMO pCBM = NewCallback();
	TDI_REQUEST r;
	TDI_STATUS err;

	if (!pCBM)
		return 0;

	r.Handle.ConnectionContext = 0;
	r.RequestNotifyObject = &SignalCallback;
	r.RequestContext = (PVOID) pCBM;
	err = pDT->TdiOpenConnectionEntry (&r, (PVOID) pSock);
	if (err == TDI_PENDING)
		err = WaitCallback(pCBM);
	if (err != TDI_SUCCESS) {
		FreeCallback(pCBM);
		return 0;
	}

	r.RequestNotifyObject = &SignalCallback;
	r.RequestContext = (PVOID) pCBM;
	err = pDT->TdiAssociateAddressEntry (&r, addr);
	if (err == TDI_PENDING)
		err = WaitCallback(pCBM);
	if (err != TDI_SUCCESS) {
		FreeCallback(pCBM);
		return 0;
	}

	FreeCallback(pCBM);
	return r.Handle.ConnectionContext;
}

int ConnectEvents = 0x10000000;
int ConnectCallbacks = 0x20000000;


PTDI_CONNECTION_INFORMATION
newTDICI (void)
{
	PTDI_CONNECTION_INFORMATION pTDI;
	PTA_ADDRESS pTAA;
	PTRANSPORT_ADDRESS pTA;

	pTDI = _HeapAllocate(sizeof(TDI_CONNECTION_INFORMATION) +
						 100, HEAPZEROINIT);
	if (!pTDI)
		return NULL;

	pTDI->RemoteAddressLength = 4 + 2 + 2 + TDI_ADDRESS_LENGTH_IP;
	pTDI->RemoteAddress = (PVOID) (pTDI + 1);

	pTA = (PTRANSPORT_ADDRESS) (pTDI + 1);
	pTA->TAAddressCount = 1;
	pTAA = (PTA_ADDRESS) &pTA->Address[0];
	pTAA->AddressLength = TDI_ADDRESS_LENGTH_IP;
	pTAA->AddressType = TDI_ADDRESS_TYPE_IP;
	return pTDI;
}

int
do_listen (PSOCKET pParent)
{
	PSOCKET pSock;
	TDI_REQUEST r;
	TDI_STATUS err;
	PTDI_CONNECTION_INFORMATION pTDIAcceptable, pTDIConnected;
	int res = 0;

	/* get a socket from the waiting list */
	_asm pushf
	_asm cli
	for (pSock = pParent->s_next; pSock; pSock = pSock->s_next) {
		switch (pSock->s_flags & (SOCKFL_LISTENUSED|SOCKFL_CONNECTED)) {
		case SOCKFL_LISTENUSED:
			/* this socket is actually listening right now; no need
			   to do anything */
			_asm popf;
			return 0;

		case 0:
			/* neither listening nor connected -- this is a socket for
			   us to grab */
			pSock->s_flags |= SOCKFL_LISTENUSED;
			goto gotit;
		}
	}
gotit:
	_asm popf

	if (!pSock)
		return 0;		/* backlog is filled */

	pTDIAcceptable = newTDICI();
	pTDIConnected = newTDICI();
	if (!pTDIAcceptable || !pTDIConnected) {
		res = -SOCKERR_NO_MEMORY;
		goto error;
	}

	if (!pSock->s_recvbuf && !(pSock->s_recvbuf = NewCBuf(5000))) {
		res = -SOCKERR_NO_MEMORY;
		goto error;
	}

	pSock->s_acceptable = pTDIAcceptable;
	pSock->s_connectinfo = pTDIConnected;
	r.Handle.ConnectionContext = pSock->s_conhdl;
	r.RequestNotifyObject = &ListenCallback;
	r.RequestContext = (PVOID) pSock;
	err = pDT->TdiListenEntry(&r, 0 & TDI_QUERY_ACCEPT, pTDIAcceptable, pTDIConnected);
	if ((err == TDI_SUCCESS) || (err == TDI_PENDING))
		pParent->s_flags |= SOCKFL_LISTENING;
	else {
		res = -err;
		goto error;
	}
	return 0;

error:
	if (pTDIAcceptable) _HeapFree (pTDIAcceptable, 0);
	if (pTDIConnected) _HeapFree (pTDIConnected, 0);
	return res;
}


void
ListenCallback (PVOID context, TDI_STATUS final, uint Bytes)
{
	PSOCKET pSock, pParent;

	pSock = (PSOCKET) context;
	if (!pSock || (pSock->s_signature != SOCKSIG))
		return;
	pParent = pSock->s_parent;
	if (!pParent || (pParent->s_signature != SOCKSIG))
		return;

	if (final != TDI_SUCCESS) {
		ConnectCallbacks |= (final << 8);
/*		pParent->s_flags &= ~SOCKFL_LISTENING; */  /*XXX this should work but causes crashes */
	} else {
		/* make current connection connected */
		pSock->s_addrobj = pParent->s_addrobj;
		AddRefAddress(pSock->s_addrobj);
		pSock->s_flags |= (SOCKFL_TCP | SOCKFL_CONNECTED);

		if (pParent->s_post) {
			(*pParent->s_post)++;
			_SignalID ((DWORD) pParent->s_post);
		}

		/* make sure we keep listening, if the backlog hasn't filled */
		do_listen(pParent);
	}
}

int
sys_listen (HVM hVM, PCRS pRegs, PSOCKET pSock)
{
	PSOCKET p, prev;
	int i, res = 0;

	/* make linked list of socket structs for listen queue */
	prev = pSock;
	for (i = 0; i < (int) pRegs->Client_ECX; i++) {
		p = NewSocket();
		if (!p) {
			res = -SOCKERR_NO_MEMORY;
			goto error;
		}
		prev->s_next = p;
		p->s_conhdl = makeconnection(pSock->s_addrobj->addr_hdl, p);
		p->s_parent = pSock;
		if (!p->s_conhdl) {
			res = -SOCKERR_INTERNAL;
			goto error;
		}
		prev = p;
	}
	prev->s_next = NULL;

	res = do_listen(pSock);
	if (res)
		goto error;
	return 0;

error:
	p = pSock->s_next;
	pSock->s_next = NULL;
	while (p) {
		prev = p;
		p = p->s_next;
		prev->s_next = NULL;
		FreeSocket(prev);
	}
	return res;
}


void
getipport (PTRANSPORT_ADDRESS pTA, PWORD pPort, PDWORD pIPAddr) 
{
	PCALLBACKMEMO pCBM = NewCallback();
	PTA_ADDRESS pTAA;
	PTDI_ADDRESS_IP pIP;

	if (pTA->TAAddressCount != 1)
		*pIPAddr = 0x01000005;
	else {
		pTAA = &pTA->Address[0];
		if (pTAA->AddressLength != TDI_ADDRESS_LENGTH_IP)
			*pIPAddr = 0x01000006;
		else if (pTAA->AddressType != TDI_ADDRESS_TYPE_IP)
			*pIPAddr = 0x01000007;
		else {
			pIP = (PTDI_ADDRESS_IP) &pTAA->Address[0];
			*pIPAddr = pIP->in_addr;
			*pPort = pIP->sin_port;
		}
	}
}
	
int
sys_accept (HVM hVM, PCRS pRegs, PSOCKET pSock)
{
#if 0
	TDI_REQUEST r;
	TDI_STATUS err;
#endif
	PCALLBACKMEMO pCBM = NewCallback();
	PSOCKET p, p2;
	int res = 0;

	if (!pCBM)
		return -SOCKERR_NO_MEMORY;

	if (!(pSock->s_flags & SOCKFL_LISTENING)) {
		res = -SOCKERR_NOT_LISTENING;
		goto error;
	}

	do {
		_asm pushf
		_asm cli
		p = pSock->s_next;
		if (!(p->s_flags & SOCKFL_CONNECTED))
			p = NULL;
		else {
			pSock->s_next = p->s_next;
			p->s_next = NULL;
		}
		_asm popf
	} while (!p && 
			 !(pSock->s_flags & SOCKFL_NONBLOCK) &&
			 (wait (pSock, POLL_READ), 1));
	/* yes, that is a use of C's little-known comma operator! */

	if (!p) {
		res = -SOCKERR_WOULD_BLOCK;
		goto error;
	}

	res = AllocateFD (hVM, pRegs->Client_EDX, &sockops, (PVOID) p);
	if (res) 
		goto error;

	p->s_sendmax = pSock->s_sendmax;
	p->s_sendlowwater = pSock->s_sendlowwater;

	getipport (p->s_connectinfo->RemoteAddress, &p->s_remoteport, &p->s_remoteip);
	pRegs->Client_EBX = p->s_remoteport;
	pRegs->Client_ECX = p->s_remoteip;

#if 0
	r.Handle.ConnectionContext = p->s_conhdl;
	r.RequestNotifyObject = &SignalCallback;
	r.RequestContext = (PVOID) pCBM;
	err = pDT->TdiAcceptEntry(&r, p->s_acceptable, p->s_connectinfo);
	if (err == TDI_PENDING)
		err = WaitCallback(pCBM);
	if (err) {
		res = -err;
		goto error;
	}
#endif

	/* make a new socket for the listen queue */
	p = NewSocket();
	if (!p)
		return -SOCKERR_NO_MEMORY;

	p->s_conhdl = makeconnection(pSock->s_addrobj->addr_hdl, p);
	if (!p->s_conhdl) {
		FreeSocket(p);
		return -SOCKERR_INTERNAL;
	}

	/* add it to queue */
	_asm pushf
	_asm cli
	for (p2 = pSock; p2; p2 = p2->s_next) {
		if (!p2->s_next) {
			p2->s_next = p;
			break;
		}
	}
	_asm popf

	res = do_listen(pSock);

error:
	if (pCBM) FreeCallback(pCBM);
	if (pSock->s_connectinfo) _HeapFree(pSock->s_connectinfo, 0);
	if (pSock->s_acceptable) _HeapFree(pSock->s_acceptable, 0);
	return res;
}

TDI_STATUS
DisconnectEvent(PVOID EventContext, PVOID ConnectionContext, uint DisconnectDataLength,
				PVOID DisconnectData, uint OptionsLength, PVOID Options, ulong Flags)
{
	PSOCKET pSock = (PSOCKET) ConnectionContext;

	if (!pSock || (pSock->s_signature != SOCKSIG))
		return TDI_SUCCESS;

	pSock->s_flags |= SOCKFL_PEERCLOSED;

	if (Flags == TDI_DISCONNECT_ABORT)
		pSock->s_flags |= SOCKFL_PEERABORTED;

	if (pSock->s_post) {
		(*pSock->s_post)++;
		_SignalID((DWORD) pSock->s_post);
	}

	return TDI_SUCCESS;
}

